package de.janicke.hj.tools;

import java.io.StreamTokenizer;
import java.io.StringReader;
import java.io.IOException;
import java.util.HashMap;

public class StreamBeautifier {
  public final static int STYLE_PASCAL = 0;
  public final static int STYLE_C = 1;

  private final static int PROCESSING_FOR = 1;
  private final static int PROCESSING_OPERATOR = 2;
  private final static int PROCESSING_STRING = 4;
  private final static int PROCESSING_ESCAPE = 8;
  private final static int PROCESSING_SWITCH = 16;
  private final static int PROCESSING_CASE = 32;
  private final static int PROCESSING_SINGLE_COMMENT = 64;
  private final static int PROCESSING_MULTI_COMMENT = 128;  

  private StreamTokenizer tokenizer = null;
  private StringBuffer out;
  private HashMap keywords = new HashMap();

  private int state = 0;
  private int roundBracketCount = 0;
  private int curlyBracketCount = 0;
  private int storeRoundBracketCount = 0;
  private int storeCurlyBracketCountSwitch = 0;
  private int lastNewLine = 0;
  private int lineCommentStart = 0;


  private int levelOfIndent = 0;
  private StringBuffer indent = new StringBuffer(80);

  private String indentPerLevel = "    ";
  private int maxCharsPerRow = 0;
  private int style = STYLE_PASCAL;
  private boolean breakStrings = false;
  private boolean breakComments = false;

  public StreamBeautifier() {
  }
  
  public void setStyle(int style) {
    this.style = style;
  }
  
  public void setIndent(String ind) {
    indentPerLevel = ind;
  }
  
  public void setBreakStrings(boolean flag) {
    breakStrings = flag;
  }

  public void setBreakComments(boolean flag) {
    breakComments = flag;
  }

  public void setBreakColumn(int col) {
    maxCharsPerRow = col;
  }

  private void setSyntaxTable() {
    tokenizer.resetSyntax();

    tokenizer.wordChars(' ', (char) 255);
    tokenizer.wordChars('\u00A0', '\u00FF');

    tokenizer.whitespaceChars('\u0000', '\u0020');

    tokenizer.ordinaryChar('(');
    tokenizer.ordinaryChar(')');
    tokenizer.ordinaryChar('{');
    tokenizer.ordinaryChar('}');
    tokenizer.ordinaryChar('[');
    tokenizer.ordinaryChar(']');
    tokenizer.ordinaryChar(';');
    tokenizer.ordinaryChar(',');
    tokenizer.ordinaryChar('=');
    tokenizer.ordinaryChar('>');
    tokenizer.ordinaryChar('<');
    tokenizer.ordinaryChar('*');
    tokenizer.ordinaryChar('+');
    tokenizer.ordinaryChar('/');
    tokenizer.ordinaryChar('-');
    tokenizer.ordinaryChar('!');
    tokenizer.ordinaryChar('^');
    tokenizer.ordinaryChar('\"');
    tokenizer.ordinaryChar('\'');
    tokenizer.ordinaryChar(':');

    tokenizer.eolIsSignificant(true);
    // tokenizer.slashSlashComments(true);

    keywords.put("if", "if");
    keywords.put("for", "for");
    keywords.put("while", "while");
    keywords.put("catch", "catch");
    keywords.put("abstract", "abstract");
    keywords.put("boolean", "boolean");
    keywords.put("break", "break");
    keywords.put("byte", "primitive");
    keywords.put("case", "case");
    keywords.put("char", "primitive");
    keywords.put("class", "class");
    keywords.put("const", "const");
    keywords.put("continue", "continue");
    keywords.put("default", "default");
    keywords.put("do", "do");
    keywords.put("double", "primitive");
    keywords.put("else", "else");
    keywords.put("extends", "extends");
    keywords.put("final", "final");
    keywords.put("float", "primitive");
    keywords.put("goto", "goto");
    keywords.put("implements", "implements");
    keywords.put("import", "import");
    keywords.put("int", "primitive");
    keywords.put("instanceof", "instanceof");
    keywords.put("interface", "forcenewline");
    keywords.put("long", "primitive");
    keywords.put("native", "native");
    keywords.put("new", "new");
    keywords.put("package", "forcenewline");
    keywords.put("private", "forcenewline");
    keywords.put("protected", "forcenewline");
    keywords.put("public", "forcenewline");
    keywords.put("return", "return");
    keywords.put("short", "primitive");
    keywords.put("static", "static");
    keywords.put("strictfp", "strictfp");
    keywords.put("super", "super");
    keywords.put("switch", "switch");
    keywords.put("synchronized", "synchronized");
    keywords.put("this", "this");
    keywords.put("throw", "throw");
    keywords.put("throws", "throws");
    keywords.put("transient", "transient");
    keywords.put("try", "try");
    keywords.put("volatile", "volatile");
    keywords.put("void", "void");
  }

  public StringBuffer beautify (String code) {
    String lastToken = "";
    String replaceValue = null;
    char quoteChar = 0;
    int strLenCounter = 0;

    out = new StringBuffer( code.length() );
    tokenizer = new StreamTokenizer( new StringReader(code) );
    setSyntaxTable();
    try {
      while (tokenizer.nextToken() != StreamTokenizer.TT_EOF) {

        if (isState(PROCESSING_STRING)) {
          // currently in String do not format!
          if ( tokenizer.ttype == quoteChar ) {
            // System.out.println("FOUND QUOTECHAR");
            if (isState(PROCESSING_ESCAPE)) {
              // was a Escape sequence code so do not end String
              clearState(PROCESSING_ESCAPE);
            } else {
              clearState(PROCESSING_STRING);
              clearState(PROCESSING_ESCAPE);
              setSyntaxTable();
              // System.out.println("END OF STRING");
            }
          } else if ( tokenizer.ttype == '\\' ) {
            if (isState(PROCESSING_ESCAPE)) {
              // System.out.println("FOUND ESCAPESEQUENCE END");
              clearState(PROCESSING_ESCAPE);
            } else {
              setState(PROCESSING_ESCAPE);
              // System.out.println("FOUND ESCAPESEQUENCE START");
            }
          } else {
            clearState(PROCESSING_ESCAPE);
          }
          
          if (breakStrings && (out.length() - lastNewLine + 3) > maxCharsPerRow) {
            // linebreak string...
            int breakindex = out.length();
            out.insert(breakindex, "\" +\n" + indent.toString() + "  \"");
            lastNewLine = breakindex; 
          }
          
          out.append((char) tokenizer.ttype);
          lastToken = (char) tokenizer.ttype + "";
          if (!isState(PROCESSING_STRING)) {
            out.append(" ");
          }
          // System.out.println("token in String = [" + lastToken + "]");
        } else if (isState(PROCESSING_SINGLE_COMMENT)) {
          // currently in single comment do not format!
          if ( tokenizer.ttype == StreamTokenizer.TT_EOL) {
              setSyntaxTable();
              clearState(PROCESSING_SINGLE_COMMENT);              
          }
          if (breakComments && (out.length() - lastNewLine) > maxCharsPerRow) {
            // linebreak single comment...
            int breakindex = out.length();
            out.insert(breakindex - 1, "\n" + indent.toString() + "// ");
            lastNewLine = breakindex; 
          }
          out.append((char) tokenizer.ttype);
          lastToken = (char) tokenizer.ttype + "";
          if (!isState(PROCESSING_SINGLE_COMMENT)) {
            // end of single line. Format...
            removeTrailingWhitespace(out);
            lastNewLine = out.length();
            out.append (indent.toString());
          }
        } else if (isState(PROCESSING_MULTI_COMMENT)) {
          // currently in multi comment do not format!
          if ( tokenizer.ttype == '/' && lastToken.equals("*")) {
              setSyntaxTable();
              clearState(PROCESSING_MULTI_COMMENT);
              out.append((char) tokenizer.ttype);
          } else if (tokenizer.ttype == StreamTokenizer.TT_EOL) {
            lastNewLine = out.length();
            out.append ("\n" + indent.toString() + indentPerLevel);
          } else if (Character.isWhitespace((char) tokenizer.ttype) 
                  && Character.isWhitespace(lastToken.charAt(0))) {
          } else {
            if (breakComments && (out.length() - lastNewLine) > maxCharsPerRow) {
              // linebreak multi comment...
              int breakindex = out.length();
              out.insert(breakindex - 1, "\n" + indent.toString() + indentPerLevel);
              lastNewLine = breakindex; 
            }
            out.append((char) tokenizer.ttype);
          }
          lastToken = (char) tokenizer.ttype + "";
          if (!isState(PROCESSING_MULTI_COMMENT)) {
            // end of multi line comment. Format...
            removeTrailingWhitespace(out);
            lastNewLine = out.length();
            out.append ("\n" + indent.toString());
          }
        } else {
          switch ( tokenizer.ttype ) {
            case StreamTokenizer.TT_WORD:
              if (isState(PROCESSING_OPERATOR)) {
                removeTrailingWhitespace(out);
                clearState(PROCESSING_OPERATOR);
              }

              // if else structure for special rules
              if (tokenizer.sval.equals("catch")) {
                removeTrailingWhitespace(out);
                out.append(tokenizer.sval + " ");
              } else if ((String) keywords.get(tokenizer.sval) == "forcenewline" ) {
                removeTrailingWhitespace(out);
                lastNewLine = out.length();
                if (!lastToken.equals("/")) out.append("\n");
                out.append("\n" + indent.toString());
                out.append(tokenizer.sval + " ");
              } else if (tokenizer.sval.equals("switch")) {
                // later replace with stack to support cascaded switch statements.
                storeCurlyBracketCountSwitch = curlyBracketCount;
                setState(PROCESSING_SWITCH);
                out.append("switch ");
              } else if (tokenizer.sval.equals("case")) {
                removeTrailingWhitespace(out);
                levelOfIndent--;
                createIndent();
                out.append("\n\n" + indent.toString() + "case ");
                setState(PROCESSING_CASE);
              } else if (tokenizer.sval.equals("else")) {
                if (lastToken.equals("}")) {
                  removeTrailingWhitespace(out);
                }
                out.append("else ");
              } else {
                out.append(tokenizer.sval + " ");
              }
              lastToken = tokenizer.sval;

              if (tokenizer.sval.equals("for")) {
                setState(PROCESSING_FOR);
              }
              break;
            case StreamTokenizer.TT_NUMBER:
              out.append(tokenizer.nval + " ");
              lastToken = tokenizer.nval + "";
              break;
            case StreamTokenizer.TT_EOL:
              // System.out.println("EOL");
              // out.append('\n');
              break;
            case '[':
            case '(':
              // check if is conditional bracket after keyword.
              replaceValue = (keywords.containsKey(lastToken) ||
                              lastToken.equals("*") ||
                              lastToken.equals("/") ||
                              lastToken.equals("+") ||
                              lastToken.equals("-") ||
                              lastToken.equals("^") ||
                              lastToken.equals("%") ||
                              lastToken.equals("&") ||
                              lastToken.equals("|") ||
                              lastToken.equals("=") 
                              ) ? (" " + (char) tokenizer.ttype) : ( (char) tokenizer.ttype + "");
              // check previous cast action!!!

              if (out.charAt(out.length()-1) == ' ' ) {
                out.replace(out.length()-1, out.length(), replaceValue);
              } else {
                out.append(replaceValue);
              }
              lastToken = (char) tokenizer.ttype + "";
              roundBracketCount++;
              break;
            case ']':
            case ')':
              if (isState(PROCESSING_OPERATOR)) {
                // there was a postfix operator so delete the blank before the operator
                if (out.charAt(out.length()-4) == ' ' ) out.deleteCharAt(out.length()-4);
              }
              clearState(PROCESSING_OPERATOR);
              if (out.charAt(out.length()-1) == ' ' ) {
                out.replace(out.length()-1, out.length(), (char) tokenizer.ttype + " ");
              } else {
                out.append((char) tokenizer.ttype + " ");
              }
              lastToken = tokenizer.ttype + "";
              roundBracketCount--;
              if (roundBracketCount == storeRoundBracketCount ) {
                clearState(PROCESSING_FOR);
              }
              break;
            case ',':
              if (isState(PROCESSING_OPERATOR)) {
                // there was a postfix operator so delete the blank before the operator
                if (out.charAt(out.length()-4) == ' ' ) out.deleteCharAt(out.length()-4);
              }
              clearState(PROCESSING_OPERATOR);
              if (out.charAt(out.length()-1) == ' ' ) {
                out.replace(out.length()-1, out.length(), ", ");
              } else {
                out.append(", ");
              }
              lastToken = (char) tokenizer.ttype + "";
              break;
            case '{':
              if (style == STYLE_PASCAL) {
                out.append('\n' + indent.toString() +  "{");
              } else if (style == STYLE_C) {
                out.append("{");
              }
              /*if (isState(PROCESSING_SWITCH) && !isState(PROCESSING_CASE)) {
                // increase indent by one for th first case xx :  only!
                levelOfIndent--;
                createIndent();
              }*/
              levelOfIndent++;
              createIndent();
              lastNewLine = out.length();
              out.append('\n' + indent.toString());
              lastToken = (char) tokenizer.ttype + "";
              curlyBracketCount++;
              break;
            case '}':
              levelOfIndent--;
              createIndent();
              /*if (isState(PROCESSING_SWITCH)) {
                levelOfIndent--;
                createIndent();
              }*/
              removeTrailingWhitespace(out);
              out.append('\n' + indent.toString() + "}\n" + indent.toString());
              lastNewLine = out.length() - indent.toString().length() - 1;
              lastToken = (char) tokenizer.ttype + "";
              curlyBracketCount--;

              if (curlyBracketCount == storeCurlyBracketCountSwitch ) {
                clearState(PROCESSING_SWITCH);
              }
              break;
            case ';':
              if (isState(PROCESSING_OPERATOR)) {
                // there was a postfix operator so delete the blank before the operator
                if (out.charAt(out.length()-4) == ' ' ) out.deleteCharAt(out.length()-4);
              }
              clearState(PROCESSING_OPERATOR);
              if (out.charAt(out.length()-1) == ' ' ) {
                out.replace(out.length()-1, out.length(), ";");
              } else {
                out.append(";");
              }
              if (!isState(PROCESSING_FOR)) {
                lastNewLine = out.length();
                out.append ("\n" + indent.toString());
              } else out.append(" ");

              lastToken = (char) tokenizer.ttype + "";
              break;
            case '\'':
            case '\"':
              if (isState(PROCESSING_STRING)) {
                // in String: not reachable case caught before switch statement
              } else {
                setState(PROCESSING_STRING);
                quoteChar = (char) tokenizer.ttype;
                tokenizer.resetSyntax();
              }
              out.append((char) tokenizer.ttype);
              lastToken = (char) tokenizer.ttype + "";
              break;
            /* case '\'':
              out.append('\'' + resetEscapeSequences(tokenizer.sval, true) + "\' ");
              lastToken = (char) tokenizer.ttype + "";
              break; */
            case '/':
              if (lastToken.equals("/")) {
                setState(PROCESSING_SINGLE_COMMENT);
                out.replace(out.length()-1, out.length(), "/");
                if (lastToken.equals("{")) {
                  removeTrailingWhitespace(out);
                } 
                tokenizer.resetSyntax();
                tokenizer.eolIsSignificant(true);
              } else {
                out.append(((char) tokenizer.ttype)+ " ");
              }
              lastToken = (char) tokenizer.ttype + "";
              break;
            case '*':
              if (lastToken.equals("/")) {
                setState(PROCESSING_MULTI_COMMENT);
                out.replace(out.length()-1, out.length(), "*");
                lastNewLine = out.length()-2;
                out.insert(out.length()-2,"\n" + indent.toString());
                lastToken = "*";
                tokenizer.resetSyntax();
              } else {
                out.append(((char) tokenizer.ttype)+ " ");
                lastToken = (char) tokenizer.ttype + "";
              }
              break;
            case '+':
              if (lastToken.equals("+")) {
                // removeTrailingWhitespace(out);
                if (out.charAt(out.length()-1) == ' ' ) out.deleteCharAt(out.length()-1);
                setState(PROCESSING_OPERATOR);
              }
              out.append("+ ");
              lastToken = "+";
              break;
            case '-':
              if (lastToken.equals("-")) {
                if (out.charAt(out.length()-1) == ' ' ) out.deleteCharAt(out.length()-1);
                setState(PROCESSING_OPERATOR);
              }
              out.append("- ");
              lastToken = "-";
              break;
            case '>':
            case '<':
              if (lastToken.equals("" + (char) tokenizer.ttype)) {
                if (out.charAt(out.length()-1) == ' ' ) out.deleteCharAt(out.length()-1);
              }
              out.append((char) tokenizer.ttype + " ");
              lastToken = "" + (char) tokenizer.ttype;
              break;
            case '=':
              // System.out.println("quation found lasttoken = [" + lastToken + "]" );
              if (lastToken.equals("+")
                  || lastToken.equals("-")
                  || lastToken.equals("*")
                  || lastToken.equals("/")
                  || lastToken.equals("^")
                  || lastToken.equals("|")
                  || lastToken.equals("&")
                  || lastToken.equals(">")
                  || lastToken.equals("<")
                  || lastToken.equals("!")
                  || lastToken.equals("=")
                  ) {
              //  System.out.println("remove whitesp.");
                if (out.charAt(out.length()-1) == ' ' ) out.deleteCharAt(out.length()-1);
              }
              out.append("= ");
              lastToken = "=";
              break;
            case ':':
              if (isState(PROCESSING_CASE)) {
                lastNewLine = out.length();
                levelOfIndent++;
                createIndent();
                out.append(":\n" + indent.toString());
                clearState(PROCESSING_CASE);
              } else {
                out.append(": ");
              }
              lastToken = ":";
              break;
            default:
              out.append(((char) tokenizer.ttype)+ " ");
              lastToken = (char) tokenizer.ttype + "";
          }

          if (maxCharsPerRow > 0 && (out.length() - lastNewLine) > maxCharsPerRow ) {
            // do a line break;
            if (lastToken.equals("\"") || lastToken.equals("\'")) {

              // it's a String, so don't break the line!
              // out.insert(out.length(), "\n" + indent.toString() + indentPerLevel + indentPerLevel);
            } else {
              int pos = findNextPossibleBreakPoint(out);
              out.insert( pos, "\n" + indent.toString() + indentPerLevel + indentPerLevel);
              lastNewLine = pos;
            }
          }
        }
      }
    } catch( IOException exc ) {
      exc.printStackTrace();
    }
    return out;
  }

  private int findNextPossibleBreakPoint(StringBuffer b) {
    int i = b.length()-2;
    char c;
    while ( i > 0 ) {
      c = b.charAt(i);
      if ( c == ' ' ) break;
      if ( c == ')' ) break;
      if ( c == ';' ) break;
      if ( c == '\'') break;
      if ( c == '\"') break;
      if ( c == '{') break;
      if ( c == '+') break;
      if ( c == '-') break;
      if ( c == '*') break;
      i--;
    }
    return i;
  }


  private void createIndent() {
    indent.delete(0, indent.length());
    for (int i=0; i<levelOfIndent; i++) indent.append(indentPerLevel);
  }


  private void removeTrailingWhitespace(StringBuffer b) {
    int end = b.length()-1;
    int start = end;
    while ( (start > 0) && (Character.isWhitespace(b.charAt(start))) ) start--;
    start++;
    if( start <= end)
      b.delete(start, end);
  }

  private void setState(int flag) {
    state |= flag;
  }

  private void clearState(int flag) {
    if (isState(flag)) state ^= flag;
  }

  private boolean isState(int flag) {
    return ((state & flag) == flag);
  }

}// end StreamBeautifier